#include <GL/glew.h>
#include <FL/Fl.H>
#include <FL/x.H>
#include <FL/fl_draw.H>
#include <FL/Fl_Gl_Window.H>
#include "gl_draw_shader.H"

#include <iostream>
using namespace std;

const char *VERTEX_SHADER_SRC =
#ifdef __APPLE__
"#version 140\n"
#else
"#version 130\n"
#endif
"\n"
"in vec4 position;\n"
"out vec2 vertexTexCoord;\n"
"uniform mat4 projection;\n"
"uniform mat4 transform;\n"
"\n"
"void main()\n"
"{\n"
"	gl_Position = projection * transform * vec4(position.xy, 0, 1);\n"
"	vertexTexCoord = position.zw;\n"
"}\n";

const char *FRAGMENT_SHADER_SRC =
#ifdef __APPLE__
"#version 140\n"
#else
"#version 130\n"
#endif
"\n"
"in vec2 vertexTexCoord;\n"
"out vec4 fragmentColor;\n"
"uniform sampler2D image;\n"
"uniform vec4 color;\n"
"\n"
"void main()\n"
"{\n"
"	float luminance = texture(image, vertexTexCoord).g;\n"
"	if (luminance == 0)\n"
"	{\n"
"		discard;\n"
"	}\n"
"	fragmentColor = vec4(color.rgb, luminance);\n"
"}\n";

Fl_Gl_Text_Static::Fl_Gl_Text_Static(Fl_Gl_Window *gl, const char *str, int x, int y, int w, int h, Fl_Align align, int draw_symbols, GLint textureNumber)
{
	_gl = gl;
	_x = x;
	_y = y;
	_w = w;
	_h = h;
	_parentWidth = _gl->w();
	_parentHeight = _gl->h();
	_textureNumber = textureNumber;
	_dataMode = GL_STATIC_DRAW;
	_fl_color = fl_color();
	_buffer = drawToTexture(str, _w, _h, align, draw_symbols);
	cout << "parent width = " << _parentWidth << ", parent height = " << _parentHeight << ", width = " << _w << ", height = " << _h << endl;
	_init = false;
}

Fl_Gl_Text_Static::~Fl_Gl_Text_Static()
{
	glUseProgram(0);
	glDeleteProgram(_program);
	glDeleteVertexArrays(1, &_vertexArray);
	glDeleteBuffers(1, &_vertexBuffer);
	glDeleteTextures(1, &_texture);
}

void Fl_Gl_Text_Static::init()
{
	if (_textureNumber < 0)
	{
		GLint maxTextures;
		glGetIntegerv(/*GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS*/GL_MAX_TEXTURE_UNITS, &maxTextures);
		_textureNumber += maxTextures;
		//_textureNumber = 1;
		cout << "_textureNumber = " << _textureNumber << endl;
	}
	_texture = createTexture(_textureNumber, _w, _h, _buffer);
	delete[] _buffer;
	_vertexBuffer = createVertexBuffer();
	_vertexArray = createVertexArray(_vertexBuffer);
	_program = createShaderProgram();
	initShaderParameters(_program, _projection, _transform, _image, _color);
	cout << "Set uniform values" << endl;
	glUseProgram(_program);
	//glActiveTexture(GL_TEXTURE0 + _textureNumber);
	//glBindTexture(GL_TEXTURE_2D, _texture);
	//glBindSampler(_textureNumber, _image);
	glUniform1i(_image, _textureNumber);
	color(_fl_color);
	GLfloat identity[] =
	{
		1, 0, 0, 0,
		0, 1, 0, 0,
		0, 0, 1, 0,
		0, 0, 0, 1
	};
	transform(identity);
	projection(_x, _y, _parentWidth, _parentHeight);
}

unsigned char *Fl_Gl_Text_Static::drawToTexture(const char *str, int &w, int &h, Fl_Align align, int draw_symbols)
{
	Fl_Color backup = fl_color();
	fl_measure(str, w, h, draw_symbols);
	cout << "Measured dimensions are: (" << w << ", " << h << ")" << endl;
	Fl_Offscreen canvas = fl_create_offscreen(w, h);
	fl_begin_offscreen(canvas);
	//fl_open_display(); // It doesn't help on X
#if defined(WIN32) || defined(__APPLE__)
	fl_rectf(0, 0, w, h, 0, 0, 0); // This crashes on X
#endif
	fl_color(255, 255, 255);
	fl_draw(str, 0, 0, w, h, align, NULL, draw_symbols);
	unsigned char *buffer = fl_read_image(0, 0, 0, w, h);
	// We need to flip the texture vertically,
	// but it is easier done in shader
	//fl_draw_image(&buffer[(h - 1) * w * 3], 0, 0, w, h, 3, -3 * w);
	//buffer = fl_read_image(0, 0, 0, w, h);
	fl_end_offscreen();
	fl_color(backup);
	return buffer;
}

GLuint Fl_Gl_Text_Static::createTexture(GLint textureNumber, int w, int h, unsigned char *buffer)
{
	GLuint texture;
	cout << "Active texture is " << (GL_TEXTURE0 + textureNumber) << endl;
	//glEnable(GL_TEXTURE_2D); - Fixed papeline
	glGenTextures(1, &texture);
	glActiveTexture((GLenum) (GL_TEXTURE0 + textureNumber));
	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	cout << "Texture image is for " << textureNumber << endl;
	glTexImage2D(GL_TEXTURE_2D, textureNumber, GL_RGB, w, h, 0, GL_RGB, GL_UNSIGNED_BYTE, buffer);
	return texture;
}

GLuint Fl_Gl_Text_Static::createVertexBuffer()
{
	// Unlike FLTK coordinates,
	// OpenGL uses bottom-up Y-axes,
	// so flip the image vertically
	// XYST vertex format X == S, Y == 1 - T
	GLfloat r = (GLfloat) _w / _h;
	GLfloat data[] =
	{
		0, 0, 0, 1,
		0, 1, 0, 0,
		r, 0, 1, 1,
		r, 1, 1, 0
	};
	GLuint vertexBuffer;
	glGenBuffers(1, &vertexBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glBufferData(GL_ARRAY_BUFFER, 4 * 4 * sizeof(GLfloat), data, _dataMode);
	return vertexBuffer;
}

GLuint Fl_Gl_Text_Static::createVertexArray(GLuint vertexBuffer)
{
	GLuint vertexArray;
	glGenVertexArrays(1, (GLuint *) &vertexArray);
	glBindVertexArray(vertexArray);
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, vertexBuffer);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, NULL);
	return vertexArray;
}

GLuint Fl_Gl_Text_Static::createShaderProgram()
{
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	const GLchar *vertexSrc[] = {VERTEX_SHADER_SRC};
	const GLchar *fragmentSrc[] = {FRAGMENT_SHADER_SRC};
	const GLuint vertex = glCreateShader(GL_VERTEX_SHADER);
	if (vertex == 0)
	{
		cout << "Error creating vertex shader" << endl;
	}
	const GLuint fragment = glCreateShader(GL_FRAGMENT_SHADER);
	if (fragment == 0)
	{
		cout << "Error creating fragment shader" << endl;
	}
	glShaderSource(vertex, 1, vertexSrc, NULL);
	glShaderSource(fragment, 1, fragmentSrc, NULL);
	glCompileShader(vertex);
	GLint statusVertex;
	glGetShaderiv(vertex, GL_COMPILE_STATUS, &statusVertex);
	if (statusVertex == GL_FALSE)
	{
		GLint len;
		glGetShaderiv(vertex, GL_INFO_LOG_LENGTH, &len);
		if (len > 0)
		{
			char *log = new char[len];
			GLsizei written;
			glGetShaderInfoLog(vertex, len, &written, log);
			cout << "Vertex shader problems:\n" << log << endl;
			delete[] log;
		}
	}
	glCompileShader(fragment);
	GLint statusFragment;
	glGetShaderiv(fragment, GL_COMPILE_STATUS, &statusFragment);
	if (statusFragment == GL_FALSE)
	{
		GLint len;
		glGetShaderiv(fragment, GL_INFO_LOG_LENGTH, &len);
		if (len > 0)
		{
			char *log = new char[len];
			GLsizei written;
			glGetShaderInfoLog(fragment, len, &written, log);
			cout << "Fragment shader problems:\n" << log << endl;
			delete[] log;
		}
	}
	const GLuint program = glCreateProgram();
	if (program == 0)
	{
		cout << "Error creating program" << endl;
	}
	glAttachShader(program, vertex);
	glAttachShader(program, fragment);
	glLinkProgram(program);
	GLint statusProgram;
	glGetProgramiv(program, GL_LINK_STATUS, &statusProgram);
	if (statusProgram == GL_FALSE)
	{
		GLint len;
		glGetProgramiv(program, GL_INFO_LOG_LENGTH, &len);
		if (len > 0)
		{
			char *log = new char[len];
			GLsizei written;
			glGetProgramInfoLog(program, len, &written, log);
			cout << "Program problems:\n" << log << endl;
			delete[] log;
		}
	}
	glDetachShader(program, vertex);
	glDetachShader(program, fragment);
	glDeleteShader(vertex);
	glDeleteShader(fragment);
	return program;
}

void Fl_Gl_Text_Static::initShaderParameters(GLuint program, GLint &projection, GLint &transform, GLint &image, GLint &color)
{
	cout << "Get uniforms locations" << endl;
	glUseProgram(program);
	projection = glGetUniformLocation(program, "projection");
	transform = glGetUniformLocation(program, "transform");
	image = glGetUniformLocation(program, "image");
	color = glGetUniformLocation(program, "color");
	cout << "program = " << program << endl;
	cout << "projection = " << projection << endl;
	cout << "transform = " << transform << endl;
	cout << "image = " << image << endl;
	cout << "color = " << color << endl;
}

void Fl_Gl_Text_Static::position(int x, int y)
{
	_x = x;
	_y = y;
	projection(_x, _y, _parentWidth, _parentHeight);
}

void Fl_Gl_Text_Static::color(uchar red, uchar green, uchar blue)
{
	cout << "Setting color to (" << (int) red << ", " << (int) green << ", " << (int) blue << ")" << endl;
	glUseProgram(_program);
	glUniform4f(_color, red / 255.0, green / 255.0, blue / 255.0, 1.0);
}

void Fl_Gl_Text_Static::color(Fl_Color c)
{
	uchar red, green, blue;
	Fl::get_color(c, red, green, blue);
	color(red, green, blue);
}

void Fl_Gl_Text_Static::projection(int x, int y, int parentWidth, int parentHeight)
{
	GLfloat sw = 2 * (GLfloat) _h / parentWidth;
	GLfloat sh = 2 * (GLfloat) _h / parentHeight;
	GLfloat dx = 2.0 * x / parentWidth - 1.0;
	GLfloat dy = 1.0 - 2.0 * (y + _h) / parentHeight;
	GLfloat matrix[] =
	{
		sw, 0, 0, 0,
		0, sh, 0, 0,
		0, 0, -1, 0,
		dx, dy, 0, 1
	};

	glUseProgram(_program);
	glUniformMatrix4fv(_projection, 1, GL_FALSE, matrix);
}

void Fl_Gl_Text_Static::transform(GLfloat *matrix)
{
	glUseProgram(_program);
	glUniformMatrix4fv(_transform, 1, GL_FALSE, matrix);
}

void Fl_Gl_Text_Static::draw()
{
	if (!_init)
	{
		init();
		_init = true;
	}
	if (_gl->w() != _parentWidth || _gl->h() != _parentHeight)
	{
		_parentWidth = _gl->w();
		_parentHeight = _gl->h();
		projection(_x, _y, _parentWidth, _parentHeight);
	}
	glUseProgram(_program);
	glActiveTexture(GL_TEXTURE0 + _textureNumber);
	glBindTexture(GL_TEXTURE_2D, _texture);
	//glBindSampler(_textureNumber, _image);
	glBindVertexArray(_vertexArray);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
}

/*
 * This function is intended to be used instead of gl_draw(...) inside Fl_Gl_Window::draw()
 * when FLTK is initialized with FL_OPENGL3 mode. Namely, under MacOS check the context mode
 * and use this function instead of existing one.
 *
 * Please note, this implementation is very inefficient, it creates the texture on each repaint.
 * For backward compatibility only
 */
void gl_draw_shader(const char *str, int x, int y, int w, int h, Fl_Align align, int draw_symbols)
{
	Fl_Gl_Window *win = Fl_Gl_Window::current()->as_gl_window();
	Fl_Gl_Text_Static text(win, str, x, y, w, h, align, draw_symbols, 0 /*-1*/);
	text._dataMode = GL_DYNAMIC_DRAW;
	text.draw();
}
